from tests.fixtures.input_fixtures import source_dir
from tests.utils import run_linter


def test_linting(source_dir):
    run_linter('core', f'{source_dir}/tests/.pylintrc', 7)
    run_linter('cli', f'{source_dir}/tests/.pylintrc', 5)



import os

os_type = os.name
# os_type = 'nt'

if os_type == 'nt':
    from PyInquirer import prompt
    type_mapping = {}
else:
    from inquirer import prompt, List, Checkbox, Confirm
    type_mapping = {
        'list': List,
        'checkbox': Checkbox,
        'confirm': Confirm
    }


__all__ = [
    'branch_selector',
    'file_selector',
    'files_checkboxes',
    'verify_prompt'
]


def interact(opt):
    def decorator(f):
        def wrapper(*args, **kwargs):
            rv = f(*args, **kwargs)
            return rv[opt] if rv not in [None, {}] else None
        return wrapper
    return decorator


def get_questions(name, message, inq_type, choices=None):
    question = {
        'message': message,
        'name': name,
        'choices': choices
    }

    if os_type == 'nt':
        question.update({'type': inq_type})
        if inq_type == 'confirm':
            question.update({'default': False})
        elif inq_type == 'checkbox':
            question['choices'] = [{'name': choice} for choice in choices]
        return [
            question
        ]
    else:
        return [
            type_mapping[inq_type](**question)
        ]


@interact('response')
def verify_prompt(message):
    questions = get_questions('response', message, 'confirm')
    return prompt(questions)


@interact('file')
def file_selector(files):
    questions = get_questions('file', 'Choose file', 'list', choices=files)
    return prompt(questions)


@interact('branch')
def branch_selector(branches):
    questions = get_questions('branch', 'Choose branch from local', 'list', choices=branches)
    return prompt(questions)


@interact('files')
def files_checkboxes(files, method):
    questions = get_questions('files', f'Choose files to {method} [space key]', 'checkbox', choices=files)
    return prompt(questions)


if __name__ == '__main__':
    res = verify_prompt('testing verify')
    print(res)

    res = file_selector(['file1', 'file2'])
    print(res)

    res = files_checkboxes(['file1', 'file2'])
    print(res)



def pytest_addoption(parser):
    parser.addoption("--source_dir", action="store", default="/Users/kobarhan/workspace/gitsy")
    parser.addoption("--test_dir", action="store", default="/Users/kobarhan/workspace/gitsy_test")



import pytest
from cli.shell import run_cmd


__all__ = [
    'source_branch',
    'target_branch',
]


@pytest.fixture()
def source_branch(test_dir):
    run_cmd(f'(cd {test_dir} && git checkout state_a)')
    return 'state_a'


@pytest.fixture()
def target_branch(test_dir):
    return 'state_b'



import os
import fire
import pytest

from cli.commands import Command
from core.gitsy import Gitsy

from tests.utils import random_str, file_name_prefix


__all__ = [
    'test_file',
    'test_files',
    'gitsy',
    'gitsy_cmd',
    'gitsy_ctx',
    'gitsy_global',
]


@pytest.fixture()
def gitsy(test_dir):
    gitsy = Gitsy(test_dir)
    return gitsy


@pytest.fixture()
def gitsy_cmd(gitsy):
    return Command(gitsy)


@pytest.fixture()
def gitsy_ctx(gitsy_cmd):
    fire.Fire(gitsy_cmd, command='init')
    yield gitsy_cmd
    fire.Fire(gitsy_cmd, command='clear')


@pytest.fixture()
def gitsy_global(gitsy_cmd, test_dir):
    yield gitsy_cmd


@pytest.fixture()
def test_file(test_dir):
    file_name = f'{file_name_prefix}_{random_str(5)}'
    open(f'{test_dir}/{file_name}', 'a').close()
    yield file_name
    os.remove(f'{test_dir}/{file_name}')


@pytest.fixture()
def test_files(test_dir):
    files = [f'{file_name_prefix}_{random_str(5)}' for _ in range(0, 5)]
    for file in files:
        open(f'{test_dir}/{file}', 'a').close()
    yield files
    for file in files:
        os.remove(f'{test_dir}/{file}')



from os import path
import fire

from cli.shell import run_cmd
from core.context import Context

from tests.fixtures.input_fixtures import test_dir, source_dir
from tests.fixtures.setup_fixtures import *


def test_init_context(gitsy_cmd, test_dir):
    fire.Fire(gitsy_cmd, command='init')
    context_branch = gitsy_cmd.gitsy.get_branch()
    print(f'{test_dir}/.gitsy/context_{context_branch}.yaml')
    assert path.exists(f'{test_dir}/.gitsy/context_{context_branch}.yaml')


def test_add_one_file_to_context(gitsy_ctx, test_file):
    cmd = gitsy_ctx
    fire.Fire(cmd, command=f'init')

    fire.Fire(cmd, command=f'add {test_file}')
    context = Context(cmd.gitsy.repo).get_context()

    assert context
    assert test_file in context['working-on']


def test_add_list_files_to_context(gitsy_ctx, test_files):
    cmd = gitsy_ctx
    fire.Fire(cmd, command=f'init')

    fire.Fire(cmd, command=f'add {" ".join(test_files)}')
    context = Context(cmd.gitsy.repo).get_context()

    assert context
    for file in test_files:
        assert file in context['working-on']


def test_change_undo(gitsy_ctx, test_file, test_dir):
    cmd = gitsy_ctx
    fire.Fire(cmd, command=f'init')

    # create a new file and stage it
    run_cmd(f'(cd {test_dir} && git add .)')

    # make changes to added file
    with open(f'{test_dir}/{test_file}', 'a') as f:
        f.write('test')

    print('\n======================================')
    # verify changes made to file
    res = run_cmd(f'(cd {test_dir} && git status)')
    assert f'modified:{test_file}' in str(res).replace(' ', '')

    fire.Fire(cmd, command=f'add {test_file}')

    print('======================================')
    # gitsy: undo changes to file and verify
    fire.Fire(cmd, command='undo all allow')
    res = run_cmd(f'(cd {test_dir} && git status)')
    assert f'modified:{test_file}' not in str(res).replace(' ', '')

    fire.Fire(cmd, command=f'clear')

    print('======================================')
    # gitsy: clean up all changes and verify
    fire.Fire(cmd, command='regret all allow')
    res = run_cmd(f'(cd {test_dir} && git status)')
    assert f'working tree clean' not in str(res).replace(' ', '')


def test_branch_selector():
    pass


def test_diff_selector():
    pass


def test_add_selector():
    pass



import os
import emoji
from emoji import EMOJI_ALIAS_UNICODE
from termcolor import colored

__all__ = ['display_list', 'display_message']

os_type = os.name

# COLORS:
# -------
# 'grey'    - not usable
# 'red'     - failure
# 'green'   - success
# 'blue'    - selectors
# 'magenta' - context
# 'yellow'  - local actions
# 'cyan'    - remote actions
# 'white'   - standard (git returns)

# ICONS:
# ------
# https://www.webfx.com/tools/emoji-cheat-sheet/

colors = {
    'saving': 'yellow',
    'saved': 'green',
    'pushed': 'green',
    'up': 'cyan',
    'undo': 'green',
    'unstaged': 'green',
    'added': 'magenta',
    'working-on': 'cyan',
    'status': 'magenta',
    'removed': 'magenta',
}


icons = {
    'saving': ':floppy_disk:',
    'saved': ':thumbsup:',
    'pushed': ':thumbsup:',
    'up': ':rocket:',
    'undo': ':tada:',
    'unstaged': ':tada:',
    'added': ':tada:',
    'working-on': ':construction_worker:',
    'status': ':bar_chart:',
    'removed': ':tada:',
}


def display_list(action, items):
    if os_type == 'nt':
        print(colored(f'{EMOJI_ALIAS_UNICODE[icons[action]]} - {action}:', colors[action]))
        # TODO: consider this for windows git-bash compatibility
        # print(colored(f'* - {action}:', colors[action]))

    else:
        print(colored(emoji.emojize(f'{icons[action]} - {action}:', use_aliases=True), colors[action]))
    for item in items:
        print(colored("\t" + item, colors[action]))


def display_message(message, color, icon):
    if os_type == 'nt':
        print(colored(f'{EMOJI_ALIAS_UNICODE[icon]} - {message}', color))
        # TODO: consider this for windows git-bash compatibility
        # print(colored(f'* - {message}', color))
    else:
        print(colored(emoji.emojize(f':{icon}: - {message}', use_aliases=True), color))


if __name__ == '__main__':

    os_type = 'nt'

    display_message('some message', 'yellow', 'thumbsup')
    items = ['1', '2', '3', '4', '5']
    display_list('up', items)



from subprocess import PIPE, Popen
from termcolor import colored


def debug(mode):
    def wrapper(g):
        def full_debug(*args, **kwargs):
            debug_sep = '**************************'
            print(colored(f'running: {args[0]}', 'magenta'))
            rv = g(*args, **kwargs)
            print(colored('done.', 'magenta'))
            print(colored(debug_sep, 'blue'))
            return rv

        def pretty_debug(*args, **kwargs):
            debug_sep = '-------------------------'
            rv = g(*args, **kwargs)
            print(colored(debug_sep, 'blue'))
            return rv

        def no_debug(*args, **kwargs):
            return g(*args, **kwargs)

        if mode == 'full':
            return full_debug
        elif mode == 'pretty':
            return pretty_debug
        elif mode is None:
            return no_debug
    return wrapper


def shell(f):
    def wrapper(*args, **kwargs):
        std_output, std_error, return_code = f(*args, **kwargs)
        error = std_error.decode().strip() if std_error else None
        output = std_output.decode().strip() if std_output else None
        if return_code != 0 and error is not None:
            print(colored(error, 'red'))
            return None, error, return_code
        else:
            for message in (error, output):
                if message not in ['', None]:
                    print(message)
            return output, None, return_code
    return wrapper


@debug(None)
@shell
def run_cmd(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE):
    p = Popen(cmd, shell=True, stdout=stdout, stdin=stdin, stderr=stderr)
    output, err = p.communicate()
    return output, err, p.returncode



import os
from os.path import join
from pathlib import Path
import copy

import yaml
from git import Repo

from cli.config import BRANCH_CONTEXT, GLOBAL_CONTEXT
from cli.display import display_message, display_list
from cli.interactive import files_checkboxes, verify_prompt

from core.gitcmd import in_gitignore, add_gitignore


def contextual(func):
    def context_wrapper(self, *args, **kwargs):
        if not os.path.exists(self.context_file):
            if verify_prompt('No context found for this branch, create one?'):
                self.init_context()
            else:
                return False
        context = self.get_context()
        res = func(self, context, *args, **kwargs)
        self._update_context(context)
        return res
    return context_wrapper


class Context:

    def __init__(self, repo) -> None:
        self.repo = repo
        self.repo_path = os.path.dirname(repo.git_dir)
        self.branch = repo.active_branch.name
        self.gitsy_path = join(self.repo_path, '.gitsy')
        self.context_file = join(self.gitsy_path, f'context_{self.branch}.yaml')

    # == CONSTRUCTORS ==========================================

    @classmethod
    def from_gitsy(cls, repo):
        return cls(repo)

    @classmethod
    def from_command(cls, cmd_path=os.getcwd()):
        repo = Repo(cmd_path, search_parent_directories=True)
        return cls(repo)

    # == END CONSTRUCTORS ==========================================

    def init_context(self):
        # TODO - prompt verify for actions:
        #  adding .gitsy dir, changing gitignore, commiting locally

        # create .gitsy dir, if needed
        gitsy_dir = join(self.repo_path, '.gitsy')
        if not os.path.isdir(gitsy_dir):
            os.mkdir(gitsy_dir)

        # add .gitsy to .gitignore, if needed
        try:
            gitignore_path = os.path.join(self.repo_path, '.gitignore')
            if not in_gitignore(gitignore_path, '.gitsy/'):
                add_gitignore(gitignore_path, '.gitsy/')
                self.repo.git.add('.gitignore')
                self.repo.git.commit('-m', 'gitsy init - update gitignore')
        except Exception as e:
            print(f'unable to add gitsy root to gitignore: {e}')

        # create context file
        if not os.path.exists(self.context_file):
            display_message(f'initiating local context for branch: {self.branch}', 'green', 'house')
            with open(self.context_file, 'w') as f:
                yaml.dump(BRANCH_CONTEXT, f)
        else:
            display_message(f'local context exists for branch: {self.branch}', 'yellow', 'house')

    def get_context(self, ignore_disabled=False):
        if os.path.exists(self.context_file):
            with open(self.context_file, 'r') as f:
                context = yaml.load(f, Loader=yaml.FullLoader)
                if ignore_disabled:
                    return context
                elif context['status'] == 'disabled':
                    global_context = copy.deepcopy(GLOBAL_CONTEXT)
                    global_context['local'] = True
                    return global_context
                else:
                    return context
        else:
            return GLOBAL_CONTEXT

    def status(self):
        """
        Prints:
            🏠 - Running in branch context: state_a
            📒 - Status: enabled
            👷 - working-on:
	                data4.txt
        """

        ctx = self.get_context()

        if ctx['type'] == 'branch':
            msg = f'Running in branch context: {self.branch}'
            icon = 'house'
        elif ctx['type'] == 'global':
            msg = f'Running in global context'
            msg += f' (disabled - {self.branch})' if ctx["local"] else ''
            icon = 'earth_africa'
        display_message(msg, 'magenta', icon)

        if ctx['type'] == 'branch':
            display_message(f'Status: {ctx["status"]}', 'green', 'ledger')
            if len(ctx["working-on"]) > 0:
                display_list('working-on', ctx['working-on'])
            else:
                display_message(f'No files in context', 'yellow', 'construction_worker')

    @contextual
    def add(self, context, files):
        if len(files) == 0:
            path = Path(self.repo_path)
            # collect all files in repo
            opt_files = [str(file.relative_to(self.repo_path))
                         for file in list(path.rglob('*'))
                         if self._file_conditions(self.repo_path, file)]
            # subtract files that are already added
            opt_files = set(opt_files) - set(context['working-on'])
            if not opt_files:
                return 'No files found to add.'
            else:
                files = files_checkboxes(list(opt_files), 'add')
        added_files = []
        for file in files:
            added_files.append(self._normalize_path(self.repo_path, file))
        added_files = set(added_files) - set(context['working-on'])
        context['working-on'] += list(added_files)
        display_list('added', added_files)

    def remove(self):
        if os.path.exists(self.context_file):
            os.remove(self.context_file)
            display_message(f'context for: {self.branch} deleted', 'green', 'ledger')
        else:
            display_message(f'context not found for: {self.branch}', 'yellow', 'ledger')

    @contextual
    def clear(self, context):
        context['working-on'].clear()
        display_message(f'context cleared', 'green', 'ledger')

    @contextual
    def rm(self, context, files):
        if len(files) == 0:
            if len(context["working-on"]) > 0:
                files = files_checkboxes(context['working-on'], 'remove from context')
            else:
                display_message(f'No files in context', 'yellow', 'construction_worker')
                return
        for file in list(files):
            context['working-on'].remove(file)
        display_list('removed', list(files))

    @contextual
    def disable(self, context):
        context['status'] = 'disabled'

    def enable(self):
        context = self.get_context(ignore_disabled=True)
        context['status'] = 'enabled'
        self._update_context(context)

    # == INTERNAL AUX METHODS ==================================

    def _update_context(self, context):
        with open(self.context_file, 'w') as f:
            yaml.dump(context, f)

    @staticmethod
    def _file_conditions(root, file):
        file_parents = [p.name for p in list(file.relative_to(root).parents)]
        return file.is_file() \
               and '.git' not in file_parents \
               and '.gitsy' not in file_parents

    @staticmethod
    def _normalize_path(repo_path, file):
        if not os.path.exists(join(repo_path, file)):
            rel_path = os.path.relpath(os.getcwd(), repo_path)
            adjusted_file = join(rel_path, file)
            return adjusted_file
        else:
            return file

    # == END INTERNAL AUX METHODS ==============================



import os
import fire


from cli.commands import Command
from core.gitsy import Gitsy


def main():
    try:
        # ========== for DEV_MODE only ==========
        if 'PC' in os.environ:
            cmd = input('enter command: ') or 'add'
            fire.Fire(Command(Gitsy('/Users/kobarhan/workspace/gitsy_test')), command=cmd)
        # =======================================
        else:
            fire.Fire(Command(Gitsy()))
    except Exception as e:
        print(e)


if __name__ == '__main__':
    main()



import fire

from cli.shell import run_cmd
from cli.config import GLOBAL_STASH_PREFIX

from tests.fixtures.input_fixtures import test_dir, source_dir
from tests.fixtures.state_fixtures import source_branch, target_branch
from tests.fixtures.setup_fixtures import *


def test_branch(gitsy_cmd, test_dir, source_branch, target_branch):
    cmd = gitsy_cmd
    # make changes at source branch and verify
    run_cmd(f'(cd {test_dir} && ./make_changes.sh 1)')
    change_list = [item.a_path for item in cmd.gitsy.repo.index.diff(None)]
    assert len(change_list) > 0

    # move to target branch, verify branch, verify stashed changes
    fire.Fire(cmd, command=f'branch {target_branch}')
    assert cmd.gitsy.get_branch() == target_branch
    assert run_cmd(f'git stash list | grep {GLOBAL_STASH_PREFIX}_{source_branch}')[2] == 0

    # make changes at target branch and verify
    run_cmd(f'(cd {test_dir} && ./make_changes.sh 2)')
    change_list = [item.a_path for item in cmd.gitsy.repo.index.diff(None)]
    assert len(change_list) > 0

    # move back to source branch, verify branch, verify stashed from target branch, verify stash stack
    fire.Fire(cmd, command=f'branch {source_branch}')
    assert cmd.gitsy.get_branch() == source_branch
    assert run_cmd(f'git stash list | grep {GLOBAL_STASH_PREFIX}_{source_branch}')[2] == 1
    assert run_cmd(f'git stash list | grep {GLOBAL_STASH_PREFIX}_{target_branch}')[2] == 0

    # verify original changes are restated
    change_list = [item.a_path for item in cmd.gitsy.repo.index.diff(None)]
    assert len(change_list) > 0



import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

package_conf = {}

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "cli", "version.py")) as f:
    exec(f.read(), package_conf)

setuptools.setup(
    name="gitsy",
    version=package_conf['__version__'],
    author="kobibarhanin",
    author_email="",
    description="Git made easy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kobibarhanin/gitsy",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "gitsy=entrypoint.entrypoint:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["attrs==19.3.0", "blessed==1.17.6", "fire==0.3.1", "gitdb==4.0.5", "GitPython==3.1.7", "importlib-metadata==1.7.0", "inquirer==2.7.0", "more-itertools==8.4.0", "packaging==20.4", "pluggy==0.13.1", "py==1.9.0", "pyparsing==2.4.7", "pytest==5.4.3", "python-editor==1.0.4", "PyYAML==5.3.1", "readchar==2.0.1", "six==1.15.0", "smmap==3.0.4", "termcolor==1.1.0", "wcwidth==0.2.5", "zipp==3.1.0", "emoji==0.5.4", "PyInquirer==1.0.3"],
    python_requires='>=3.7',
)



GLOBAL_STASH_PREFIX = 'gitsy_global_stash'

DEFAULT_COMMIT = 'fast commit'

BRANCH_CONTEXT = {
    'type': 'branch',
    'status': 'enabled',
    'working-on': []
}

GLOBAL_CONTEXT = {
    'type': 'global',
    'local': False
}



import os
from pathlib import Path
from cli.config import DEFAULT_COMMIT


class Command:

    def __init__(self, gitsy):
        self.gitsy = gitsy

    #  == Git operations commands ========

    def branch(self, branch_name=None):
        """
        Switch to another branch.
        :param branch_name: (optional) Branch name to switch to.
        """
        return self.gitsy.branch(branch_name)

    def up(self, message=DEFAULT_COMMIT):
        """
        Adds, Commits and Pushed changes to remote.
        :param message: (optional) Commit message.
        """
        return self.gitsy.up(message)

    def save(self, message=DEFAULT_COMMIT):
        """
        Adds and Commits changes.
        :param message: (optional) Commit message.
        """
        return self.gitsy.save(message)

    def diff(self):
        """
        Prints diff of selected file.
        """
        return self.gitsy.diff()

    def undo(self, *scope):
        """
        Undo un-staged (non added) changes.
        :param optional file list as scope:
        """
        return self.gitsy.undo(scope)

    def regret(self, *scope):
        """
        Undo staged (added) changes.
        :param optional file list as scope:
        """
        return self.gitsy.regret(scope)

    def rename(self, name):
        """
        Rename current branch.
        :param required new branch name:
        """
        return self.gitsy.rename(name)

    def ignore(self, opt):
        """
        Enable gitignore modifications: add, remove.
        Syncs gitignore to remote.
        """
        return self.gitsy.ignore(opt)

    def test(self):
        """
        Route for dev purposes
        """
        return self.gitsy.test()

    # == Context operations commands ========

    def init(self):
        """
        Creates .gitsy directory if not yet created
        Adds .gitsy path to gitignore if not yet added
        Creates context file for current branch.
        """
        return self.gitsy.context.init_context()

    def context(self):
        """
        Prints current context information.
        """
        return self.gitsy.context.status()

    def add(self, *files):
        return self.gitsy.context.add(files)

    def rm(self, *files):
        return self.gitsy.context.rm(files)

    def clear(self):
        return self.gitsy.context.clear()

    def remove(self):
        return self.gitsy.context.remove()

    def disable(self):
        return self.gitsy.context.disable()

    def enable(self):
        return self.gitsy.context.enable()

    # == General commands ===================

    @staticmethod
    def version():
        here = Path(__file__).parent.absolute()
        package_conf = {}
        with open(os.path.join(here, "version.py")) as f:
            exec(f.read(), package_conf)
        return package_conf['__version__']



import os

from git import Repo
from git import InvalidGitRepositoryError
from git import GitCommandError
from termcolor import colored

from cli.shell import run_cmd
from cli.display import *
from cli.interactive import *
from cli.config import *

from core.context import Context
from core.config import DIFF_MAP


class Gitsy:

    def __init__(self, repo_path=os.getcwd()):
        try:
            # initialize repo object
            self.repo = Repo(repo_path, search_parent_directories=True)
            self.repo_path = os.path.dirname(self.repo.git_dir)

            # initialize attributes for methods usage
            self._cwd = os.getcwd()
            self._branch = self.repo.active_branch.name

            self.context = Context.from_gitsy(self.repo)

        except InvalidGitRepositoryError:
            raise Exception('Not a git repo, I have no power here...')

    def get_branch(self):
        return self._branch

    # ====================== NON CONTEXTUAL METHODS =====================

    def branch(self, target_branch):
        """
        Swith to target_branch if specified, else prompt branch menu.
        Implements auto-stashing to allow flex branch hopping.
        :param target_branch:
        :return:
        """
        if not target_branch:
            branches = [branch.name for branch in list(self.repo.branches)]
            if len(branches) == 1:
                return 'No local branches detected'
            else:
                target_branch = branch_selector(branches)
                if not target_branch:
                    return

        change_list = self._calculate_change_list(GLOBAL_CONTEXT,
                                                  diff_types=['unstaged', 'staged'])

        if len(change_list) > 0:
            self.repo.git.stash(f'save', f'{GLOBAL_STASH_PREFIX}_{self.get_branch()}')
            display_message(f'Saving diff (stash): {GLOBAL_STASH_PREFIX}_{self.get_branch()}', 'green', 'thumbsup')
        self._switch_branch(target_branch)

        stash_list = self.repo.git.stash('list')
        stash_stub = f'{GLOBAL_STASH_PREFIX}_{self.get_branch()}'
        if stash_list:
            stash_list = stash_list.split("\n")

        for stash in stash_list:
            stash_name = stash.split(' ')[-1]
            if stash_stub == stash_name:
                stash_index = stash.split(' ')[0].replace(':', '')
                self.repo.git.stash(f'pop', stash_index)
                display_message(f'Loading diff (stash): {stash_name}', 'green', 'thumbsup')
                break

    def rename(self, name):
        try:
            self.repo.active_branch.rename(name)
            context = self.context.get_context(ignore_disabled=True)
            if context['type'] == 'branch':
                os.rename(self.context.context_file,
                          os.path.join(self.context.gitsy_path, f'context_{name}.yaml'))
        except Exception as e:
            print(e)

    # TODO - add support for adding to gitignore (in the correct way)
    #  removing from gitignore, and syncing to remote
    @staticmethod
    def ignore(opt):
        if opt == 'reset':
            # self.repo.index.remove('.', '-r')
            run_cmd('git rm -r --cached .')
            run_cmd('git add .')
            run_cmd('git commit -m ".gitignore fixed"')

    # ====================== OPTIONALLY CONTEXTUAL METHODS =====================

    def up(self, message):
        context = self.context.get_context()
        change_list = self._calculate_change_list(context,
                                                  diff_types=['unstaged', 'staged'])
        if change_list:
            self._up(change_list, message)
        else:
            ctx = 'global' if context['type'] == 'global' else self.get_branch()
            msg = f'Nothing to send up within context: {ctx}'
            display_message(msg, 'yellow', 'speak_no_evil')

    def save(self, message):
        context = self.context.get_context()
        change_list = self._calculate_change_list(context,
                                                  diff_types=['unstaged'])
        if change_list:
            self._save(change_list, message)
        else:
            return f'No changes to save within context: {self._branch}'

    def diff(self):
        context = self.context.get_context()
        change_list = self._calculate_change_list(context,
                                                  diff_types=['unstaged'])
        if not change_list:
            return 'No diff for you!'
        file = file_selector(change_list)
        if not file:
            return
        diff = self.repo.git.diff(file).split('\n')
        for line in diff:
            if line.startswith('+++') or line.startswith('---'):
                print(colored(line, 'white'))
            elif line.startswith('+'):
                print(colored(line, 'green'))
            elif line.startswith('-'):
                print(colored(line, 'red'))
            else:
                print(colored(line, 'white'))

    def undo(self, scope):
        context = self.context.get_context()
        try:
            scope = self._get_scope(context, scope, 'undo', diff_types=['unstaged'])
        except Exception as e:
            return e
        if scope:
            self.repo.git.checkout('--', list(scope))
            display_list('undo', scope)

    def regret(self, scope):
        context = self.context.get_context()
        try:
            scope = self._get_scope(context, scope, 'regret', diff_types=['staged'])
        except Exception as e:
            return e
        self.repo.git.reset('--', scope)
        display_list('unstaged', scope)

    # ==================== INTERNAL AUX METHODS ====================

    def _up(self, files, message):
        try:
            self._save(files, message)
            display_list('up', files)
        except GitCommandError:
            print(colored(f'pushing without changes', 'cyan'))
        self.repo.git.push('origin', self._branch)
        display_message('pushed', 'green', 'thumbsup')

    def _save(self, files, message):
        display_list('saving', files)
        self.repo.git.add(files)
        commit = self.repo.git.commit('-m', message)
        display_message(commit, 'white', 'cat')
        display_message('saved', 'green', 'thumbsup')

    def _switch_branch(self, branch_name):
        self.repo.git.checkout(branch_name)
        self._branch = self.repo.active_branch.name
        display_message(f'Switched to branch: {self._branch}', 'yellow', 'checkered_flag')

    def _get_scope(self, context, scope, method, diff_types):
        change_list = self._calculate_change_list(context, diff_types)
        if len(scope) == 0:
            if len(change_list) > 0:
                scope = files_checkboxes(change_list, method)
            else:
                raise Exception(f'No actionable items in found context: {self._branch}')
        elif scope[0] == 'all':
            if len(scope) < 2 or not scope[1] == 'allow':
                if not verify_prompt(f'This will {method} all staged changes'):
                    return
            scope = change_list
        else:
            scope = [value for value in scope
                     if value in change_list]
        return scope

    def _calculate_change_list(self, context, diff_types=None):
        change_list = []
        for diff_type in diff_types:
            change_list += [item.a_path for item
                            in self.repo.index.diff(DIFF_MAP[diff_type])]
        change_list += self.repo.untracked_files
        context_list = context['working-on'] if context['type'] == 'branch' else change_list
        return self._intersection(change_list, context_list)

    @staticmethod
    def _intersection(iterable1, iterable2):
        return list(set(iterable1).intersection(iterable2))

    # ==================== TEST METHODS ====================

    def test(self):
        print('=== TEST ROUTE ===')
        print(self.branch)
        print('=== END TEST ROUTE ===')




def in_gitignore(gitignore_path, item):
    gitignore = []
    with open(gitignore_path, 'r') as ignore_file:
        for line in ignore_file.readlines():
            gitignore.append(line.strip())
    return item in gitignore


def add_gitignore(gitignore_path, item):
    with open(gitignore_path, 'a') as ignore_file:
        ignore_file.write(item)